# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
import json
from ..aws.pagination import build_result
from ..aws.services import (
    extract_pagination_config,
)
from ..common.command import IRCommand, OutputFile
from ..common.config import (
    AWS_MAX_ATTEMPTS,
    CONNECT_TIMEOUT_SECONDS,
    READ_TIMEOUT_SECONDS,
    get_user_agent_extra,
)
from ..common.file_system_controls import validate_file_path
from ..common.helpers import Boto3Encoder, operation_timer
from botocore.config import Config
from jmespath.parser import ParsedResult
from typing import Any


TIMEOUT_AFTER_SECONDS = 10
CHUNK_SIZE = 4 * 1024 * 1024


def interpret(
    ir: IRCommand,
    access_key_id: str,
    secret_access_key: str,
    session_token: str | None,
    region: str,
    client_side_filter: ParsedResult | None = None,
    max_results: int | None = None,
    endpoint_url: str | None = None,
) -> dict[str, Any]:
    """Interpret the given intermediate representation into boto3 calls.

    The function returns the response from the operation indicated by the
    intermediate representation.
    """
    config_result = extract_pagination_config(ir.parameters, max_results)
    parameters = config_result.parameters
    pagination_config = config_result.pagination_config

    config = Config(
        region_name=region,
        connect_timeout=CONNECT_TIMEOUT_SECONDS,
        read_timeout=READ_TIMEOUT_SECONDS,
        retries={'total_max_attempts': AWS_MAX_ATTEMPTS, 'mode': 'adaptive'},
        user_agent_extra=get_user_agent_extra(),
    )

    with operation_timer(ir.service_name, ir.operation_python_name, region):
        client = boto3.client(
            ir.service_name,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            aws_session_token=session_token,
            config=config,
            endpoint_url=endpoint_url,
        )

        if client.can_paginate(ir.operation_python_name):
            response = build_result(
                paginator=client.get_paginator(ir.operation_python_name),
                service_name=ir.service_name,
                operation_name=ir.operation_name,
                operation_parameters=ir.parameters,
                pagination_config=pagination_config,
                client_side_filter=client_side_filter,
            )
        else:
            operation = getattr(client, ir.operation_python_name)
            response = operation(**parameters)

            if client_side_filter is not None:
                response = _apply_filter(response, client_side_filter)

        if ir.has_streaming_output and ir.output_file and ir.output_file.path != '-':
            response = _handle_streaming_output(response, ir.output_file)

        return response


def _handle_streaming_output(response: dict[str, Any], output_file: OutputFile) -> dict[str, Any]:
    streaming_output = response[output_file.response_key]

    # Validate file path before writing
    validated_path = validate_file_path(output_file.path)

    with open(validated_path, 'wb') as f:
        for chunk in streaming_output.iter_chunks(chunk_size=CHUNK_SIZE):
            f.write(chunk)

    del response[output_file.response_key]
    return response


def _apply_filter(response: dict[str, Any], client_side_filter: ParsedResult) -> dict[str, Any]:
    response_metadata = response.get('ResponseMetadata')
    json_compatible_response = json.loads(json.dumps(response, cls=Boto3Encoder))
    filtered_result = client_side_filter.search(json_compatible_response)
    return {'Result': filtered_result, 'ResponseMetadata': response_metadata}



import pytest
from awslabs.aws_api_mcp_server.core.metadata.read_only_operations_list import (
    DEFAULT_REQUEST_TIMEOUT,
    SERVICE_REFERENCE_URL,
    ReadOnlyOperations,
    ServiceReferenceUrlsByService,
)
from requests import Response
from unittest.mock import MagicMock, call, patch


TEST_SERVICE = 'testService'
TEST_URL = 'https://test-url.json'
TEST_READ_OPERATION = 'TestReadOperation'
TEST_READ_OPERATION_2 = 'TestReadOperation2'
TEST_WRITE_OPERATION = 'TestWriteOperation'


@pytest.fixture
def sample_service_reference_list_response():
    """Fixture providing a sample policy document."""
    return [{'service': TEST_SERVICE, 'url': TEST_URL}]


@pytest.fixture
def sample_service_reference_response():
    """Fixture providing a sample policy document."""
    return {
        'Name': TEST_SERVICE,
        'Actions': [
            {
                'Name': TEST_READ_OPERATION,
                'ActionConditionKeys': [],
                'Annotations': {
                    'Properties': {
                        'IsList': False,
                        'IsPermissionManagement': False,
                        'IsTaggingOnly': False,
                        'IsWrite': False,
                    }
                },
                'Resources': [{'Name': TEST_READ_OPERATION}],
            },
            {
                'Name': TEST_READ_OPERATION_2,
                'ActionConditionKeys': [],
                'Annotations': {
                    'Properties': {
                        'IsList': False,
                        'IsPermissionManagement': False,
                        'IsTaggingOnly': False,
                        'IsWrite': False,
                    }
                },
                'Resources': [{'Name': TEST_READ_OPERATION_2}],
            },
            {
                'Name': TEST_WRITE_OPERATION,
                'ActionConditionKeys': [],
                'Annotations': {
                    'Properties': {
                        'IsList': False,
                        'IsPermissionManagement': False,
                        'IsTaggingOnly': False,
                        'IsWrite': True,
                    }
                },
                'Resources': [{'Name': TEST_WRITE_OPERATION}],
            },
        ],
    }


@patch('requests.get')
def test_read_only_operations_initialization(
    mocked_requests_get, sample_service_reference_list_response
):
    """Test ReadOnlyOperations initialization."""
    mocked_service_reference_list_response = MagicMock(spec=Response)
    mocked_service_reference_list_response.json.return_value = (
        sample_service_reference_list_response
    )
    mocked_requests_get.return_value = mocked_service_reference_list_response

    operations = ReadOnlyOperations(ServiceReferenceUrlsByService())

    assert isinstance(operations, dict)
    mocked_requests_get.assert_called_once_with(
        SERVICE_REFERENCE_URL, timeout=DEFAULT_REQUEST_TIMEOUT
    )


@patch('requests.get')
def test_read_only_operations_has_method_missing_service(
    mocked_requests_get, sample_service_reference_list_response, sample_service_reference_response
):
    """Test the has method of ReadOnlyOperations when the provided service is missing."""
    mocked_service_reference_list_response = MagicMock(spec=Response)
    mocked_service_reference_list_response.json.return_value = (
        sample_service_reference_list_response
    )
    mocked_service_reference_response = MagicMock(spec=Response)
    mocked_service_reference_response.json.return_value = sample_service_reference_response
    mocked_requests_get.side_effect = [
        mocked_service_reference_list_response,
        mocked_service_reference_response,
    ]

    operations = ReadOnlyOperations(ServiceReferenceUrlsByService())

    assert isinstance(operations, dict)
    assert operations.has(TEST_SERVICE, TEST_READ_OPERATION)
    assert not operations.has(TEST_SERVICE, TEST_WRITE_OPERATION)
    mocked_requests_get.assert_has_calls(
        [
            call(SERVICE_REFERENCE_URL, timeout=DEFAULT_REQUEST_TIMEOUT),
            call(TEST_URL, timeout=DEFAULT_REQUEST_TIMEOUT),
        ],
        any_order=False,
    )


@patch('requests.get')
def test_read_only_operations_has_method_second_call_for_service_queries_local_cache(
    mocked_requests_get, sample_service_reference_list_response, sample_service_reference_response
):
    """Test the has method of ReadOnlyOperations when the provided service is available."""
    mocked_service_reference_list_response = MagicMock(spec=Response)
    mocked_service_reference_list_response.json.return_value = (
        sample_service_reference_list_response
    )
    mocked_service_reference_response = MagicMock(spec=Response)
    mocked_service_reference_response.json.return_value = sample_service_reference_response
    mocked_requests_get.side_effect = [
        mocked_service_reference_list_response,
        mocked_service_reference_response,
    ]

    operations = ReadOnlyOperations(ServiceReferenceUrlsByService())

    assert isinstance(operations, dict)
    # First call for a service, should get data from service reference API
    assert operations.has(TEST_SERVICE, TEST_READ_OPERATION)
    # Second call for the same service, should lookup data from local cache
    assert operations.has(TEST_SERVICE, TEST_READ_OPERATION_2)
    mocked_requests_get.assert_has_calls(
        [
            call(SERVICE_REFERENCE_URL, timeout=DEFAULT_REQUEST_TIMEOUT),
            call(TEST_URL, timeout=DEFAULT_REQUEST_TIMEOUT),
        ],
        any_order=False,
    )
    assert mocked_requests_get.call_count == 2


@patch('requests.get')
def test_read_only_operations_has_method_error(
    mocked_requests_get, sample_service_reference_list_response
):
    """Test the has method of ReadOnlyOperations when the service reference API call throws an error."""
    mocked_response = MagicMock(spec=Response)
    mocked_response.json.return_value = sample_service_reference_list_response
    mocked_requests_get.side_effect = [
        mocked_response,
        RuntimeError('Error while calling service reference API'),
    ]

    operations = ReadOnlyOperations(ServiceReferenceUrlsByService())

    assert isinstance(operations, dict)
    with pytest.raises(RuntimeError):
        operations.has(TEST_SERVICE, TEST_READ_OPERATION)
    mocked_requests_get.assert_has_calls(
        [
            call(SERVICE_REFERENCE_URL, timeout=DEFAULT_REQUEST_TIMEOUT),
            call(TEST_URL, timeout=DEFAULT_REQUEST_TIMEOUT),
        ],
        any_order=False,
    )


@patch('requests.get')
def test_service_reference_urls_by_service_error(mocked_requests_get):
    """Test ServiceReferenceUrlsByService initialization when the service reference API call throws an error."""
    mocked_requests_get.side_effect = RuntimeError('Error while calling service reference API')

    with pytest.raises(RuntimeError):
        ServiceReferenceUrlsByService()
    mocked_requests_get.assert_has_calls(
        [call(SERVICE_REFERENCE_URL, timeout=DEFAULT_REQUEST_TIMEOUT)], any_order=False
    )


def test_read_only_operations_has_method_custom_operation():
    """Test the has method of ReadOnlyOperations with custom operations."""
    operations = ReadOnlyOperations({})
    assert operations.has('s3', 'ls')
    assert operations.has('logs', 'start-live-tail')
    assert not operations.has('s3', 'sync')


def test_read_only_operations_has_method_operation_from_metadata():
    """Test the has method of ReadOnlyOperations with operations defined in api metadata."""
    operations = ReadOnlyOperations({})
    assert operations.has('s3', 'ListBuckets')
    assert operations.has('lambda', 'ListAliases')
    assert operations.has('rds', 'DescribeDBSnapshotAttributes')
    assert not operations.has('s3', 'DeleteObject')
    assert not operations.has('lambda', 'CreateAlias')
    assert not operations.has('rds', 'CreateDBSecurityGroup')


def test_read_only_operations_overrides():
    """Test the has method of ReadOnlyOperations with overrides."""
    operations = ReadOnlyOperations({})
    assert not operations.has('sts', 'AssumeRole')
    assert not operations.has('sts', 'AssumeRoleWithWebIdentity')
    assert not operations.has('sts', 'AssumeRoleWithSAML')
    assert not operations.has('sts', 'GetSessionToken')
    assert not operations.has('sts', 'GetFederationToken')
    assert not operations.has('sts', 'AssumeRoot')
    assert not operations.has('iam', 'CreateAccessKey')
    assert not operations.has('cognito-identity', 'GetCredentialsForIdentity')
    assert not operations.has('cognito-identity', 'GetOpenIdToken')
    assert not operations.has('sso', 'GetRoleCredentials')



import pytest
from awslabs.aws_api_mcp_server.core.common.help_command import (
    _clean_description,
    _clean_text,
    generate_help_document,
)
from unittest.mock import MagicMock, patch


class MockDoc:
    """Mock class for ReSTDocument."""

    def __init__(self, content=b''):
        """Initialize the mock document."""
        self.content = content

    def getvalue(self):
        """Get the value of the document."""
        return self.content


def test_clean_text():
    """Test cleaning text."""
    assert _clean_text('  hello   world  ') == 'hello world'
    assert _clean_text('hello\nworld') == 'hello world'


def test_clean_description():
    """Test cleaning description."""
    desc = '=== Description ===\n\nThis is a description.'
    assert _clean_description(desc) == 'This is a description.'

    desc = 'This is a description.'
    assert _clean_description(desc) == 'This is a description.'


@patch('awslabs.aws_api_mcp_server.core.common.help_command.driver')
def test_generate_help_document_unknown_command(mock_driver):
    """Test generating help document for unknown command."""
    service_name = 'unknown'
    operation_name = 'op'

    mock_command_table = MagicMock()
    mock_driver._get_command_table.return_value = mock_command_table

    mock_command_table.__getitem__.return_value = (
        MagicMock()
    )  # Neither BasicCommand nor ServiceCommand

    result = generate_help_document(service_name, operation_name)
    assert result is None


@pytest.mark.parametrize(
    'service_name, operation_name, expected_text, expected_params',
    [
        ('s3', 'ls', 'List S3 objects and common prefixes', ['recursive', 'page-size']),
        (
            'ec2',
            'describe-instances',
            'Describes the specified instances',
            ['filters', 'instance-ids'],
        ),
        ('lambda', 'list-functions', 'Returns a list of Lambda functions', ['max-items']),
        ('sts', 'get-caller-identity', 'Returns details about the IAM user', []),
        ('dynamodb', 'list-tables', 'Returns an array of table names', ['max-items']),
        ('iam', 'get-user', 'Retrieves information about the specified IAM user', ['user-name']),
        ('sns', 'publish', 'Sends a message to an Amazon SNS topic', ['topic-arn', 'message']),
        (
            'sqs',
            'send-message',
            'Delivers a message to the specified queue',
            ['queue-url', 'message-body'],
        ),
    ],
)
def test_generate_help_document_real_aws_commands(
    service_name, operation_name, expected_text, expected_params
):
    """Test generating help documents for real AWS CLI commands."""
    result = generate_help_document(service_name, operation_name)

    assert result is not None
    assert result['command'] == f'aws {service_name} {operation_name}'
    assert expected_text in result['description']

    for param in expected_params:
        assert param in result['parameters']
